/* if.c
**
** This is the `shared object' version of the database interface; it
** requires the programming interface to the dynimic loader - namely `libdl.a'
** and `dlfcn.h'.
*/
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <dlfcn.h>

#include "dbif.h"
#include "tools.h"
#include "vxerror.h"
#include "prefs.h"
#include "version.h"
#include "mrmacs.h"

typedef void (*reset_ft) (DBIF_FILE *, char *);
typedef void (*create_ft) (DBIF_FILE *, char *, int);
typedef void (*rw_ft) (DBIF_FILE *, char *);
typedef void (*close_ft) (DBIF_FILE);
typedef void (*sync_ft) (DBIF_FILE);
typedef dbif_rec *(*resolve_ft) (DBIF_FILE, char *, char **, int);
typedef dbif_rec *(*get_ft) (DBIF_FILE, char *);
typedef int (*insert_ft) (DBIF_FILE, char *, dbif_rec *);
typedef int (*set_ft) (DBIF_FILE, char *, dbif_rec *);
typedef int (*del_ft) (DBIF_FILE, char *);
typedef const char *(*errormsg_ft) (void);
typedef dbif_error (*errno_ft) (void);
typedef dbif_rec *(*mkrecl_ft)(char*,size_t,char*,size_t,size_t,off_t,int,...);
typedef dbif_rec *(*mkrec_ft) (char *, char *, size_t, off_t, int, ...);
typedef int (*rectype_ft) (const dbif_rec *);
typedef int (*setlim_ft) (dbif_rec *, int, int);
typedef int (*getlim_ft) (dbif_rec *, int *, int *);
typedef void (*free_ft) (dbif_rec *);
typedef DBIF_CURSOR (*select_ft) (DBIF_FILE, char *, seltype);
typedef char *(*first_ft) (DBIF_CURSOR, dbif_rec **);
typedef char *(*next_ft) (DBIF_CURSOR, dbif_rec **);
typedef int (*end_ft) (DBIF_CURSOR);

static void *dbif_lhdl = NULL;
static struct dbif_funclist {
    reset_ft idbif_reset;
    create_ft idbif_create;
    rw_ft idbif_rw;
    close_ft idbif_close;
    sync_ft idbif_sync;
    resolve_ft idbif_resolve;
    get_ft idbif_get;
    insert_ft idbif_insert;
    set_ft idbif_set;
    del_ft idbif_del;
    errormsg_ft idbif_errormsg;
    errno_ft idbif_errno;
    mkrecl_ft idbif_mkrecl;
    mkrec_ft idbif_mkrec;
    rectype_ft idbif_rectype;
    setlim_ft idbif_setlim;
    getlim_ft idbif_getlim;
    free_ft idbif_free;
    select_ft idbif_select;
    first_ft idbif_first;
    next_ft idbif_next;
    end_ft idbif_end;
} if_funcs;

const char *(*idbif_identity)(void), *(*idbif_version)(void);
void (*idbif_copyprefs)(struct preferences *oprefs);

static const char *dbif_dlerror (void)
{
    char *err, *e1;
    ifnull (e1 = dlerror ()) { return NULL; }
    err = strchr (e1, ':');
    if (err) ++err; else err = e1;
    return err;
}

static void dbif_init_check (char *file, int line, char *symname)
{
    char *err = (char *) dbif_dlerror ();
    unlessnull (err) {
	vxerror_at (EX_OSERR, 0, file, line,
		    "symbol `%s' not found in module: %s", symname, err);
    }
}

#define check_set(type, var,sym) \
	    do { \
		var = (type) dlsym(dbif_lhdl, (sym)); \
		dbif_init_check (__FILE__, __LINE__, (sym)); \
	    } while (0);


static void dbif_init_module (void)
{
    FILE **ilogfile;
    int *iloglevel;
    char **iprogname;
    unlessnull (dbif_lhdl) { return; }
    dbif_lhdl = dlopen (prefs.dbif_module, RTLD_NOW);
    ifnull (dbif_lhdl) {
	vxerror_at (EX_OSERR, 0, __FILE__, __LINE__ - 2,
		    "i couldn't open the VxMail DBIF module:\n\t%s\n",
		    dbif_dlerror ());
    }

    /* First consistency check */
    check_set (const char *(*)(void), idbif_identity, "dbif_modid");
    if (strcmp (idbif_identity(), "VxMail DBIF loadable module")) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "invalid VxMail DBIF module!\n");
    }

    /* Second consistency check */
    check_set (const char *(*)(void), idbif_version, "dbif_modver");
    if (strcmp (idbif_version(), vxVersion ())) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "wrong version in VxMail DBIF module!\n");
    }

    /* load the global data from the main program into the if-module... */
    ilogfile = dlsym (dbif_lhdl, "logfile");		/* logfile */
    unlessnull (ilogfile) *ilogfile = logfile;
    iloglevel = dlsym (dbif_lhdl, "loglevel");		/* loglevel */
    unlessnull (iloglevel) *iloglevel = loglevel;
    iprogname = dlsym (dbif_lhdl, "progname");		/* program name */
    unlessnull (iprogname) *iprogname = progname;

    /* Copy the preferences to the module's `prefs' structure */
    check_set (void(*)(struct preferences *),idbif_copyprefs, "dbif_copyprefs");
    idbif_copyprefs (&prefs);

    /* Initialize the function pointers */
    check_set (reset_ft, if_funcs.idbif_reset, "dbif_reset");
    check_set (create_ft, if_funcs.idbif_create, "dbif_create");
    check_set (rw_ft, if_funcs.idbif_rw, "dbif_rw");
    check_set (close_ft, if_funcs.idbif_close, "dbif_close");
    check_set (sync_ft, if_funcs.idbif_sync, "dbif_sync");
    check_set (resolve_ft, if_funcs.idbif_resolve, "dbif_resolve");
    check_set (get_ft, if_funcs.idbif_get, "dbif_get");
    check_set (insert_ft, if_funcs.idbif_insert, "dbif_insert");
    check_set (set_ft, if_funcs.idbif_set, "dbif_set");
    check_set (del_ft, if_funcs.idbif_del, "dbif_del");
    check_set (errormsg_ft, if_funcs.idbif_errormsg, "dbif_errormsg");
    check_set (errno_ft, if_funcs.idbif_errno, "dbif_errno");
    check_set (mkrecl_ft, if_funcs.idbif_mkrecl, "dbif_mkrecl");
    check_set (mkrec_ft, if_funcs.idbif_mkrec, "dbif_mkrec");
    check_set (rectype_ft, if_funcs.idbif_rectype, "dbif_rectype");
    check_set (setlim_ft, if_funcs.idbif_setlim, "dbif_setlim");
    check_set (getlim_ft, if_funcs.idbif_getlim, "dbif_getlim");
    check_set (free_ft, if_funcs.idbif_free, "dbif_free");
    check_set (select_ft, if_funcs.idbif_select, "dbif_select");
    check_set (first_ft, if_funcs.idbif_first, "dbif_first");
    check_set (next_ft, if_funcs.idbif_next, "dbif_next");
    check_set (end_ft, if_funcs.idbif_end, "dbif_end");
}

void dbif_reset (DBIF_FILE *dbf, char *dbname)
{
    dbif_init_module ();
    if_funcs.idbif_reset (dbf, dbname);
}

void dbif_create (DBIF_FILE *dbf, char *dbname, int mode)
{
    dbif_init_module ();
    if_funcs.idbif_create (dbf, dbname, mode);
}

void dbif_rw (DBIF_FILE *dbf, char *dbname)
{
    dbif_init_module ();
    if_funcs.idbif_rw (dbf, dbname);
}

void dbif_close (DBIF_FILE dbf)
{
    dbif_init_module ();
    if_funcs.idbif_close (dbf);
}

void dbif_sync (DBIF_FILE dbf)
{
    dbif_init_module ();
    if_funcs.idbif_sync (dbf);
}

dbif_rec *dbif_resolve (DBIF_FILE db, char *em_addr, char **res_addr, int iur)
{
    dbif_init_module ();
    return if_funcs.idbif_resolve (db, em_addr, res_addr, iur);
}

dbif_rec *dbif_get (DBIF_FILE db, char *em_addr)
{
    dbif_init_module ();
    return if_funcs.idbif_get (db, em_addr);
}

int dbif_insert (DBIF_FILE db, char *em_addr, dbif_rec *rec)
{
    dbif_init_module ();
    return if_funcs.idbif_insert (db, em_addr, rec);
}

int dbif_set (DBIF_FILE db, char *em_addr, dbif_rec *rec)
{
    dbif_init_module ();
    return if_funcs.idbif_set (db, em_addr, rec);
}

int dbif_del (DBIF_FILE db, char *em_addr)
{
    dbif_init_module ();
    return if_funcs.idbif_del (db, em_addr);
}

const char *dbif_errormsg (void)
{
    dbif_init_module ();
    return if_funcs.idbif_errormsg ();
}

dbif_error dbif_errno (void)
{
    dbif_init_module ();
    return if_funcs.idbif_errno ();
}

dbif_rec *dbif_mkrecl (char *passwd, size_t pwl,
		       char *folder, size_t fll,
		       size_t msgmax, off_t fldmax,
		       int rtype,
		       ...)
{
    va_list ap;
    int nusers, maxusers;
    dbif_init_module ();
    if (rtype == RT_DOMAIN || rtype == RT_NSDOMAIN) {
	va_start (ap, rtype);
	nusers = va_arg (ap, int);
	maxusers = va_arg (ap, int);
	va_end (ap);
	return if_funcs.idbif_mkrecl (passwd, pwl, folder, fll,
				      msgmax, fldmax,
				      rtype, nusers, maxusers);
    } else {
	return if_funcs.idbif_mkrecl (passwd, pwl, folder, fll,
				      msgmax, fldmax,
				      rtype);
    }
}

dbif_rec *dbif_mkrec (char *passwd, char *folder,
		      size_t msgmax, off_t fldmax,
		      int rtype, ...)
{
    va_list ap;
    int nusers, maxusers;
    dbif_init_module ();
    if (rtype == RT_DOMAIN || rtype == RT_NSDOMAIN) {
	va_start (ap, rtype);
	nusers = va_arg (ap, int);
	maxusers = va_arg (ap, int);
	va_end (ap);
	return if_funcs.idbif_mkrec (passwd, folder,
				     msgmax, fldmax,
				     rtype, nusers, maxusers);
    } else {
	return if_funcs.idbif_mkrec (passwd, folder,
				     msgmax, fldmax,
				     rtype);
    }
}

int dbif_rectype (const dbif_rec *r)
{
    dbif_init_module ();
    return if_funcs.idbif_rectype (r);
}

int dbif_setlim (dbif_rec *r, int nusers, int maxusers)
{
    dbif_init_module ();
    return if_funcs.idbif_setlim (r, nusers, maxusers);
}

int dbif_getlim (dbif_rec *r, int *nusers, int *maxusers)
{
    dbif_init_module ();
    return if_funcs.idbif_getlim (r, nusers, maxusers);
}

void dbif_free (dbif_rec *p)
{
    dbif_init_module ();
    if_funcs.idbif_free (p);
}

DBIF_CURSOR dbif_select (DBIF_FILE db, char *key, seltype sel)
{
    dbif_init_module ();
    return if_funcs.idbif_select (db, key, sel);
}

char *dbif_first (DBIF_CURSOR dbc, dbif_rec **userrec)
{
    dbif_init_module ();
    return if_funcs.idbif_first (dbc, userrec);
}

char *dbif_next (DBIF_CURSOR dbc, dbif_rec **userrec)
{
    dbif_init_module ();
    return if_funcs.idbif_next (dbc, userrec);
}

int dbif_end (DBIF_CURSOR dbc)
{
    dbif_init_module ();
    return if_funcs.idbif_end (dbc);
}

void dbif_loadif (void)
{
    if (prefs.loadif_at == DBIFATSERV) {
	vxlog (4, "loading %s interface", iftype());
	dbif_init_module ();
	vxlog (4, " done.\n");
    }
}

void dbif_unloadif (void)
{
    ifnull (dbif_lhdl) { return; }
    vxlog (4, "un-loading %s interface", iftype());
    if (dlclose (dbif_lhdl)) {
	vxerror_at (EX_SOFTWARE, errno, __FILE__, __LINE__, "unloadif failed");
    }
    dbif_lhdl = NULL;
    vxlog (4, " done.\n");
}
